@@ -20,24 +20,24 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <sys/time.h> /* For the timing of dvdcss_title crack. */
-#include <fcntl.h>
-#include <stdlib.h>
-#include <stdio.h>
-#include <errno.h>
-#include <string.h>
-#include <ctype.h>
-#include <unistd.h>
-#include <limits.h>
-#include <dirent.h>
+#include "config.h"
+#include <sys/types.h>      /* off_t */
+#include <sys/stat.h>       /* stat */
+#include <sys/time.h>       /* For the timing of dvdcss_title crack. */
+#include <fcntl.h>          /* open */
+#include <stdlib.h>         /* free */
+#include <stdio.h>          /* fprintf */
+#include <errno.h>          /* errno, EIN* */
+#include <string.h>         /* memcpy, strlen */
+#include <unistd.h>         /* chdir, getcwd */
+#include <limits.h>         /* PATH_MAX */
+#include <dirent.h>         /* opendir, readdir */
 
 /* misc win32 helpers */
 #ifdef WIN32
-#ifndef HAVE_GETTIMEOFDAY
-/* replacement gettimeofday implementation */
-#include <sys/timeb.h>
+# ifndef HAVE_GETTIMEOFDAY
+   /* replacement gettimeofday implementation */
+#  include <sys/timeb.h>
 static inline int _private_gettimeofday( struct timeval *tv, void *tz )
 {
   struct timeb t;
@@ -46,29 +46,35 @@ static inline int _private_gettimeofday( struct timeval *tv, void *tz )
   tv->tv_usec = t.millitm * 1000;
   return 0;
 }
-#define gettimeofday(TV, TZ) _private_gettimeofday((TV), (TZ))
-#endif
-#include <io.h> /* read() */
-#define lseek64 _lseeki64
+#  define gettimeofday(TV, TZ) _private_gettimeofday((TV), (TZ))
+# endif
+# include <io.h> /* read() */
+# define lseek64 _lseeki64
 #endif
 
 #if defined(__FreeBSD__) || defined(__OpenBSD__) || defined(__NetBSD__) || defined(__bsdi__) || defined(__APPLE__)
-#define SYS_BSD 1
+# define SYS_BSD 1
 #endif
 
 #if defined(__sun)
-#include <sys/mnttab.h>
+# include <sys/mnttab.h>
+#elif defined(__APPLE__)
+# include <sys/param.h>
+# include <sys/ucred.h>
+# include <sys/mount.h>
 #elif defined(SYS_BSD)
-#include <fstab.h>
+# include <fstab.h>
 #elif defined(__linux__)
-#include <mntent.h>
-#include <paths.h>
+# include <mntent.h>
+# include <paths.h>
 #endif
 
 #include "dvdread/dvd_udf.h"
-#include "dvd_input.h"
 #include "dvdread/dvd_reader.h"
+#include "dvd_input.h"
+#include "dvdread_internal.h"
 #include "md5.h"
+#include "dvdread/ifo_read.h"
 
 #define DEFAULT_UDF_CACHE_LEVEL 1
 
@@ -113,7 +119,7 @@ struct dvd_file_s {
   ssize_t filesize;
 };
 
-int UDFReadBlocksRaw( dvd_reader_t *device, uint32_t lb_number,
+int InternalUDFReadBlocksRaw( const dvd_reader_t *device, uint32_t lb_number,
                       size_t block_count, unsigned char *data,
                       int encrypted );
 
@@ -235,11 +241,12 @@ static dvd_reader_t *DVDOpenImageFile( const char *location, int have_css )
     return NULL;
   }
 
-  dvd = (dvd_reader_t *) malloc( sizeof( dvd_reader_t ) );
+  dvd = malloc( sizeof( dvd_reader_t ) );
   if( !dvd ) {
     dvdinput_close(dev);
     return NULL;
   }
+  memset( dvd, 0, sizeof( dvd_reader_t ) );
   dvd->isImageFile = 1;
   dvd->dev = dev;
   dvd->path_root = NULL;
@@ -263,14 +270,14 @@ static dvd_reader_t *DVDOpenPath( const char *path_root )
 {
   dvd_reader_t *dvd;
 
-  dvd = (dvd_reader_t *) malloc( sizeof( dvd_reader_t ) );
+  dvd = malloc( sizeof( dvd_reader_t ) );
   if( !dvd ) return NULL;
   dvd->isImageFile = 0;
   dvd->dev = 0;
   dvd->path_root = strdup( path_root );
   if(!dvd->path_root) {
     free(dvd);
-    return 0;
+    return NULL;
   }
   dvd->udfcache_level = DEFAULT_UDF_CACHE_LEVEL;
   dvd->udfcache = NULL;
@@ -294,6 +301,7 @@ static char *sun_block2char( const char *path )
 
   /* Replace "/dsk/" with "/rdsk/" */
   new_path = malloc( strlen(path) + 2 );
+  if(!new_path) return NULL;
   strcpy( new_path, path );
   strcpy( strstr( new_path, "/dsk/" ), "" );
   strcat( new_path, "/rdsk/" );
@@ -325,6 +333,7 @@ static char *bsd_block2char( const char *path )
 
   /* Replace "/dev/" with "/dev/r" */
   new_path = malloc( strlen(path) + 2 );
+  if(!new_path) return NULL;
   strcpy( new_path, "/dev/r" );
   strcat( new_path, path + strlen( "/dev/" ) );
 
@@ -400,6 +409,8 @@ dvd_reader_t *DVDOpen( const char *ppath )
 #else
     dev_name = strdup( path );
 #endif
+    if(!dev_name)
+        goto DVDOpen_error;
     dvd = DVDOpenImageFile( dev_name, have_css );
     free( dev_name );
     free(path);
@@ -474,7 +485,28 @@ dvd_reader_t *DVDOpen( const char *ppath )
       path_copy[1] = '\0';
     }
 
-#if defined(SYS_BSD)
+#if defined(__APPLE__)
+    struct statfs s[128];
+    int r = getfsstat(NULL, 0, MNT_NOWAIT);
+    if (r > 0) {
+        if (r > 128)
+            r = 128;
+        r = getfsstat(s, r * sizeof(s[0]), MNT_NOWAIT);
+        int i;
+        for (i=0; i<r; i++) {
+            if (!strcmp(path_copy, s[i].f_mntonname)) {
+                dev_name = bsd_block2char(s[i].f_mntfromname);
+                fprintf( stderr,
+                        "libdvdread: Attempting to use device %s"
+                        " mounted on %s for CSS authentication\n",
+                        dev_name,
+                        s[i].f_mntonname);
+                auth_drive = DVDOpenImageFile( dev_name, have_css );
+                break;
+            }
+        }
+    }
+#elif defined(SYS_BSD)
     if( ( fe = getfsfile( path_copy ) ) ) {
       dev_name = bsd_block2char( fe->fs_spec );
       fprintf( stderr,
@@ -601,13 +633,13 @@ static dvd_file_t *DVDOpenFileUDF( dvd_reader_t *dvd, char *filename )
 
   start = UDFFindFile( dvd, filename, &len );
   if( !start ) {
-    fprintf( stderr, "libdvdnav:DVDOpenFileUDF:UDFFindFile %s failed\n", filename );
+    fprintf( stderr, "libdvdread:DVDOpenFileUDF:UDFFindFile %s failed\n", filename );
     return NULL;
   }
 
-  dvd_file = (dvd_file_t *) malloc( sizeof( dvd_file_t ) );
+  dvd_file = malloc( sizeof( dvd_file_t ) );
   if( !dvd_file ) {
-    fprintf( stderr, "libdvdnav:DVDOpenFileUDF:malloc failed\n" );
+    fprintf( stderr, "libdvdread:DVDOpenFileUDF:malloc failed\n" );
     return NULL;
   }
   dvd_file->dvd = dvd;
@@ -690,19 +722,19 @@ static dvd_file_t *DVDOpenFilePath( dvd_reader_t *dvd, char *filename )
 
   /* Get the full path of the file. */
   if( !findDVDFile( dvd, filename, full_path ) ) {
-    fprintf( stderr, "libdvdnav:DVDOpenFilePath:findDVDFile %s failed\n", filename );
+    fprintf( stderr, "libdvdread:DVDOpenFilePath:findDVDFile %s failed\n", filename );
     return NULL;
   }
 
   dev = dvdinput_open( full_path );
   if( !dev ) {
-    fprintf( stderr, "libdvdnav:DVDOpenFilePath:dvdinput_open %s failed\n", full_path );
+    fprintf( stderr, "libdvdread:DVDOpenFilePath:dvdinput_open %s failed\n", full_path );
     return NULL;
   }
 
-  dvd_file = (dvd_file_t *) malloc( sizeof( dvd_file_t ) );
+  dvd_file = malloc( sizeof( dvd_file_t ) );
   if( !dvd_file ) {
-    fprintf( stderr, "libdvdnav:DVDOpenFilePath:dvd_file malloc failed\n" );
+    fprintf( stderr, "libdvdread:DVDOpenFilePath:dvd_file malloc failed\n" );
     dvdinput_close(dev);
     return NULL;
   }
@@ -716,6 +748,7 @@ static dvd_file_t *DVDOpenFilePath( dvd_reader_t *dvd, char *filename )
   if( stat( full_path, &fileinfo ) < 0 ) {
     fprintf( stderr, "libdvdread: Can't stat() %s.\n", filename );
     free( dvd_file );
+    dvdinput_close( dev );
     return NULL;
   }
   dvd_file->title_sizes[ 0 ] = fileinfo.st_size / DVD_VIDEO_LB_LEN;
@@ -739,7 +772,7 @@ static dvd_file_t *DVDOpenVOBUDF( dvd_reader_t *dvd, int title, int menu )
   start = UDFFindFile( dvd, filename, &len );
   if( start == 0 ) return NULL;
 
-  dvd_file = (dvd_file_t *) malloc( sizeof( dvd_file_t ) );
+  dvd_file = malloc( sizeof( dvd_file_t ) );
   if( !dvd_file ) return NULL;
   dvd_file->dvd = dvd;
   /*Hack*/ dvd_file->css_title = title << 1 | menu;
@@ -782,7 +815,7 @@ static dvd_file_t *DVDOpenVOBPath( dvd_reader_t *dvd, int title, int menu )
   dvd_file_t *dvd_file;
   int i;
 
-  dvd_file = (dvd_file_t *) malloc( sizeof( dvd_file_t ) );
+  dvd_file = malloc( sizeof( dvd_file_t ) );
   if( !dvd_file ) return NULL;
   dvd_file->dvd = dvd;
   /*Hack*/ dvd_file->css_title = title << 1 | menu;
@@ -881,7 +914,7 @@ dvd_file_t *DVDOpenFile( dvd_reader_t *dvd, int titlenum,
     }
     break;
   case DVD_READ_TITLE_VOBS:
-    if( titlenum == 0 ) return 0;
+    if( titlenum == 0 ) return NULL;
     if( dvd->isImageFile ) {
       return DVDOpenVOBUDF( dvd, titlenum, 0 );
     } else {
@@ -904,7 +937,7 @@ void DVDCloseFile( dvd_file_t *dvd_file )
 {
   int i;
 
-  if( dvd_file ) {
+  if( dvd_file && dvd_file->dvd ) {
     if( !dvd_file->dvd->isImageFile ) {
       for( i = 0; i < TITLES_MAX; ++i ) {
         if( dvd_file->title_devs[ i ] ) {
@@ -914,7 +947,7 @@ void DVDCloseFile( dvd_file_t *dvd_file )
     }
 
     free( dvd_file );
-    dvd_file = 0;
+    dvd_file = NULL;
   }
 }
 
@@ -1093,7 +1126,7 @@ int DVDFileStat( dvd_reader_t *dvd, int titlenum,
 }
 
 /* Internal, but used from dvd_udf.c */
-int UDFReadBlocksRaw( dvd_reader_t *device, uint32_t lb_number,
+int InternalUDFReadBlocksRaw( const dvd_reader_t *device, uint32_t lb_number,
                       size_t block_count, unsigned char *data,
                       int encrypted )
 {
@@ -1121,11 +1154,11 @@ int UDFReadBlocksRaw( dvd_reader_t *device, uint32_t lb_number,
  * into the buffer located at 'data' and if 'encrypted' is set
  * descramble the data if it's encrypted.  Returning either an
  * negative error or the number of blocks read. */
-static int DVDReadBlocksUDF( dvd_file_t *dvd_file, uint32_t offset,
+static int DVDReadBlocksUDF( const dvd_file_t *dvd_file, uint32_t offset,
                              size_t block_count, unsigned char *data,
                              int encrypted )
 {
-  return UDFReadBlocksRaw( dvd_file->dvd, dvd_file->lb_start + offset,
+  return InternalUDFReadBlocksRaw( dvd_file->dvd, dvd_file->lb_start + offset,
                            block_count, data, encrypted );
 }
 
@@ -1135,7 +1168,7 @@ static int DVDReadBlocksUDF( dvd_file_t *dvd_file, uint32_t offset,
  * into the buffer located at 'data' and if 'encrypted' is set
  * descramble the data if it's encrypted.  Returning either an
  * negative error or the number of blocks read. */
-static int DVDReadBlocksPath( dvd_file_t *dvd_file, unsigned int offset,
+static int DVDReadBlocksPath( const dvd_file_t *dvd_file, unsigned int offset,
                               size_t block_count, unsigned char *data,
                               int encrypted )
 {
@@ -1287,7 +1320,7 @@ ssize_t DVDReadBytes( dvd_file_t *dvd_file, void *data, size_t byte_size )
   numsec = ( ( seek_byte + byte_size ) / DVD_VIDEO_LB_LEN ) +
     ( ( ( seek_byte + byte_size ) % DVD_VIDEO_LB_LEN ) ? 1 : 0 );
 
-  secbuf_base = (unsigned char *) malloc( numsec * DVD_VIDEO_LB_LEN + 2048 );
+  secbuf_base = malloc( numsec * DVD_VIDEO_LB_LEN + 2048 );
   secbuf = (unsigned char *)(((uintptr_t)secbuf_base & ~((uintptr_t)2047)) + 2048);
   if( !secbuf_base ) {
     fprintf( stderr, "libdvdread: Can't allocate memory "
@@ -1328,20 +1361,35 @@ int DVDDiscID( dvd_reader_t *dvd, unsigned char *discid )
 {
   struct md5_ctx ctx;
   int title;
+  int title_sets;
   int nr_of_files = 0;
+  ifo_handle_t *vmg_ifo;
 
   /* Check arguments. */
   if( dvd == NULL || discid == NULL )
     return 0;
 
-  /* Go through the first 10 IFO:s, in order,
+  vmg_ifo = ifoOpen( dvd, 0 );
+  if( !vmg_ifo ) {
+    fprintf( stderr, "libdvdread: DVDDiscId, failed to "
+      "open VMG IFO!\n" );
+    return -1;
+  }
+
+  title_sets = vmg_ifo->vmgi_mat->vmg_nr_of_title_sets + 1;
+  ifoClose( vmg_ifo );
+
+  if( title_sets > 10 )
+  	title_sets = 10;
+
+  /* Go through the first IFO:s, in order, up until the tenth,
    * and md5sum them, i.e  VIDEO_TS.IFO and VTS_0?_0.IFO */
   md5_init_ctx( &ctx );
-  for( title = 0; title < 10; title++ ) {
+  for( title = 0; title < title_sets; title++ ) {
     dvd_file_t *dvd_file = DVDOpenFile( dvd, title, DVD_READ_INFO_FILE );
     if( dvd_file != NULL ) {
       ssize_t bytes_read;
-      size_t file_size = dvd_file->filesize * DVD_VIDEO_LB_LEN;
+      ssize_t file_size = dvd_file->filesize * DVD_VIDEO_LB_LEN;
       char *buffer_base = malloc( file_size + 2048 );
       char *buffer = (char *)(((uintptr_t)buffer_base & ~((uintptr_t)2047)) + 2048);
 
@@ -1401,7 +1449,7 @@ int DVDISOVolumeInfo( dvd_reader_t *dvd,
     return -1;
   }
 
-  ret = UDFReadBlocksRaw( dvd, 16, 1, buffer, 0 );
+  ret = InternalUDFReadBlocksRaw( dvd, 16, 1, buffer, 0 );
   if( ret != 1 ) {
     fprintf( stderr, "libdvdread: DVDISOVolumeInfo, failed to "
              "read ISO9660 Primary Volume Descriptor!\n" );
